import pandas as pd
import numpy as np

import lwd.core.decorators

import lwd.data.resolution

import lwd.analysis.patterns
import lwd.analysis.forecast

@lwd.core.decorators.class_methods
@pd.api.extensions.register_dataframe_accessor('lwd')
class DataFrameToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.patterns = self.patterns(self._obj)
        self.forecast = self.forecast(self._obj)

    @lwd.core.decorators.class_methods
    class patterns():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        # patterns
        def polynomial(self, column=None, **kwargs):
            """
            Find the polynomial of a series and project a bandwidth
                arguments
                series: pandas.Series
                shape: int or tuple
                    when an int is provided the polynomial is provided as n-th degree polynomial
                    when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
                    default value: (0, 6)
                validation: string
                    validation type for shape tuple
                    p_val: test for normal distribution of the residuals
                    rsq: check for improvement of the rsq value
                    default value: p_val
                threshold: float
                    the threshold for normal distribution test or rsq improvement
                    default value: 0.05
                confidence: float
                    the bound confidence
                    default value: 0.8
                outlier_sigma: float
                    outliers are datapoints outside the outlier_sigma fraction
                    default value: 2
                adjust: boolean
                    adjust polynomial for outliers
                    default value: True
            """
            if isinstance(column, str) and (column in self._obj):
                return lwd.analysis.patterns.Polynomial(self._obj[column], **kwargs)
            else:
                print(column)
                print(self._obj)
                raise ValueError(f"Please provide a valid column name to analyse.")

        def periodical(self, column=None, **kwargs):
            """
            Find the periodical pattern of a series and project a bandwidth
                series: pandas.Series
                rule: period representation used for resampling the series
                    default value: "1W"
                how: define how the period must be evaluated
                    options are "mean", "min", "max" and "std"
                    default value: "mean"
                resolution: define the resolution of the pattern
                    the pattern is rounded to fit the resolution
                    default value: None
                confidence: float
                    the bandwidth confidence
                    default value: 0.8
                outlier_sigma: float
                    outliers are datapoints outside the outlier_sigma fraction
                    default value: 2    
            """
            if isinstance(column, str) and column in self._obj:
                return lwd.analysis.patterns.Periodical(self._obj[column], **kwargs)
            else:
                raise ValueError(f"Please provide a valid column name to analyse.")

    @lwd.core.decorators.class_methods
    class forecast():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        def MAPE(self, **kwargs):
            """
            Return the mean absolute percentage error.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.MAPE(self._obj, **kwargs)

        def SMAPE(self, **kwargs):
            """
            Return the symmetric mean absolute percentage error.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.SMAPE(self._obj, **kwargs)
        
        def MDA(self, **kwargs):
            """
            Return the mean mean directional accuracy.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.MDA(self._obj, **kwargs)

        def MPE(self, **kwargs):
            """
            Return the mean percentage error.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.MPE(self._obj, **kwargs)

        mape = MAPE
        smape = SMAPE
        mda = MDA
        mpe = MPE