import pandas as pd
import numpy as np

import lwd.core.decorators

import lwd.data.resolution

import lwd.analysis.patterns
import lwd.analysis.forecast

@pd.api.extensions.register_series_accessor('lwd')
class SeriesToolkit():
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self.data = self.data(self._obj)
        self.patterns = self.patterns(self._obj)
        self.forecast = self.forecast(self._obj)
        
    @lwd.core.decorators.class_methods
    class data():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        def group_index(self, **kwargs):
            """
            Group a series by DateTime index and apply a specific function.
                arguments
                series: pandas.Series
                index: str
                    a string with date-time keywords that can be parsed to group the index
                    keywords:
                        weekday_name -> %a  # abbreviated weekday name
                        week_day -> %u      # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                        weekday -> %u       # weekday as a number (1 to 7), Monday=1. Warning: In Sun Solaris Sunday=1
                        iso_week -> %V      # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                        isoweek -> %V       # The ISO 8601 week number of the current year (01 to 53), where week 1 is the first week that has at least 4 days in the current year, and with Monday as the first day of the week
                        monthname -> %b     # abbreviated month name
                        month_name -> %b    # abbreviated month name
                        monthday -> %d      # day of the month (01 to 31)
                        month_day -> %d     # day of the month (01 to 31)
                        isoyear -> %G       # 4-digit year corresponding to the ISO week number (see %V).
                        iso_year -> %G      # 4-digit year corresponding to the ISO week number (see %V).
                        yearday -> %j       # day of the year (001 to 366)
                        year_day -> %j      # day of the year (001 to 366)
                        year -> %Y          # year including the century
                        month -> %m         # month (01 to 12)
                        week -> %W          # week number of the current year, starting with the first Monday as the first day of the first week
                        day -> %d           # day of the month (01 to 31)
                        hour -> %H          # hour, using a 24-hour clock (00 to 23)
                        minute -> %M        # minute
                        second -> %S        # second
                        time -> %T          # current time, equal to %H:%M:%S
                    
                    default value "iso_year-iso_week"

                    e.g. "iso_year-Wiso_week" will parse to "%G-W%V" which will print "2021-W01" for the first week of 2021

                Apply one of the following functions:
                    .sum()
                    .mean()
                    .min()
                    .max()
                    .std()
                    .value_range()
                    .count()
                    .subsize_count()
                    .subseries_count()

                e.g. series.blue.data.group_index(rule="1W").sum()
                
            """
            return lwd.data.resolution.GroupByDatetimeIndex(self._obj, **kwargs)
        
    @lwd.core.decorators.class_methods
    class patterns():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        # patterns
        def polynomial(self, **kwargs):
            """
            Find the polynomial of a series and project a bandwidth
                arguments
                series: pandas.Series
                shape: int or tuple
                    when an int is provided the polynomial is provided as n-th degree polynomial
                    when a tuple is provided the function will find an optimised polynomial between first and second value of the tuple
                    default value: (0, 6)
                validation: string
                    validation type for shape tuple
                    p_val: test for normal distribution of the residuals
                    rsq: check for improvement of the rsq value
                    default value: p_val
                threshold: float
                    the threshold for normal distribution test or rsq improvement
                    default value: 0.05
                confidence: float
                    the bound confidence
                    default value: 0.8
                outlier_sigma: float
                    outliers are datapoints outside the outlier_sigma fraction
                    default value: 2
                adjust: boolean
                    adjust polynomial for outliers
                    default value: True
            """
            return lwd.analysis.patterns.Polynomial(self._obj, **kwargs)

        def periodical(self, **kwargs):
            """
            Find the periodical pattern of a series and project a bandwidth
                series: pandas.Series
                rule: period representation used for resampling the series
                    default value: "1W"
                how: define how the period must be evaluated
                    options are "mean", "min", "max" and "std"
                    default value: "mean"
                resolution: define the resolution of the pattern
                    the pattern is rounded to fit the resolution
                    default value: None
                confidence: float
                    the bandwidth confidence
                    default value: 0.8
                outlier_sigma: float
                    outliers are datapoints outside the outlier_sigma fraction
                    default value: 2    
            """
            return lwd.analysis.patterns.Periodical(self._obj, **kwargs)
    
    @lwd.core.decorators.class_methods
    class forecast():

        def __init__(self, pandas_obj):
            self._obj = pandas_obj
                
        def MAPE(self, **kwargs):
            """
            Return the mean absolute percentage error.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.MAPE(self._obj, **kwargs)

        def SMAPE(self, **kwargs):
            """
            Return the symmetric mean absolute percentage error.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.SMAPE(self._obj, **kwargs)

        def MDA(self, **kwargs):
            """
            Return the mean mean directional accuracy.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.MDA(self._obj, **kwargs)

        def MPE(self, **kwargs):
            """
            Return the mean percentage error.
                arguments
                ef __init__(self, frame, forecast=None, actuals=None, **kwargs):
                frame: pandas.Series or pandas.DataFrame
                forecast: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the forecast columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with forecast data; frame will be treated as actuals data
                    default value None
                actuals: pandas.Series or str
                    if frame is a pandas.DataFrame this must be a string with the actuals columns name                    
                    if frame is a pandas.Series this can be a pandas.Series with actuals data; frame will be treated as forecast data
                    default value None
                if frame is a pandas.Series only one of forecast and actuals must be provided
            """
            return lwd.analysis.forecast.MPE(self._obj, **kwargs)

        mape = MAPE
        smape = SMAPE
        mda = MDA
        mpe = MPE